@@ -1,6 +1,5 @@
-﻿"""
 """
-from __future__ import absolute_import
+"""
 
 import copy
 from functools import wraps
@@ -10,8 +9,11 @@ import numpy as np
 from . import markup
 from .dimensionality import Dimensionality, p_dict
 from .registry import unit_registry
-from .utilities import with_doc
+from .decorators import with_doc
 
+PREFERRED = []  # List of preferred quantities for each symbol,
+                # e.g. PREFERRED = [pq.mV, pq.pA, pq.UnitQuantity('femtocoulomb', 1e-15*pq.C, 'fC')]
+                # Intended to be overwritten in down-stream packages
 
 def validate_unit_quantity(value):
     try:
@@ -27,7 +29,10 @@ def validate_unit_quantity(value):
 
 def validate_dimensionality(value):
     if isinstance(value, str):
-        return unit_registry[value].dimensionality
+        try:
+            return unit_registry[value].dimensionality
+        except (KeyError, UnicodeDecodeError):
+            return unit_registry[str(value)].dimensionality
     elif isinstance(value, Quantity):
         validate_unit_quantity(value)
         return value.dimensionality
@@ -50,10 +55,11 @@ def get_conversion_factor(from_u, to_u):
 def scale_other_units(f):
     @wraps(f)
     def g(self, other, *args):
+        other = np.asanyarray(other)
         if not isinstance(other, Quantity):
-            other = Quantity(other, copy=False)
+            other = other.view(type=Quantity)
         if other._dimensionality != self._dimensionality:
-            other = other.rescale(self.units)
+            other = other.rescale(self.units, dtype=np.result_type(self.dtype, other.dtype))
         return f(self, other, *args)
     return g
 
@@ -109,10 +115,12 @@ class Quantity(np.ndarray):
     __array_priority__ = 21
 
     def __new__(cls, data, units='', dtype=None, copy=True):
-        if isinstance(data, cls):
+        if isinstance(data, Quantity):
             if units:
                 data = data.rescale(units)
-            return np.array(data, dtype=dtype, copy=copy, subok=True)
+            if isinstance(data, unit_registry['UnitQuantity']):
+                return 1*data
+            return np.array(data, dtype=dtype, copy=copy, subok=True).view(cls)
 
         ret = np.array(data, dtype=dtype, copy=copy).view(cls)
         ret._dimensionality.update(validate_dimensionality(units))
@@ -126,7 +134,7 @@ class Quantity(np.ndarray):
     def _reference(self):
         """The reference quantity used to perform conversions"""
         rq = 1*unit_registry['dimensionless']
-        for u, d in self.dimensionality.iteritems():
+        for u, d in self.dimensionality.items():
             rq = rq * u._reference**d
         return rq * self.magnitude
 
@@ -134,16 +142,34 @@ class Quantity(np.ndarray):
     def magnitude(self):
         return self.view(type=np.ndarray)
 
+    @property
+    def real(self):
+        return Quantity(self.magnitude.real, self.dimensionality)
+
+    @real.setter
+    def real(self, r):
+        self.magnitude.real = Quantity(r, self.dimensionality).magnitude
+
+    @property
+    def imag(self):
+        return Quantity(self.magnitude.imag, self.dimensionality)
+
+    @imag.setter
+    def imag(self, i):
+        self.magnitude.imag = Quantity(i, self.dimensionality).magnitude
+
     @property
     def simplified(self):
         rq = 1*unit_registry['dimensionless']
-        for u, d in self.dimensionality.iteritems():
+        for u, d in self.dimensionality.items():
             rq = rq * u.simplified**d
         return rq * self.magnitude
 
-    def _get_units(self):
-        return Quantity(1.0, self.dimensionality)
-    def _set_units(self, units):
+    @property
+    def units(self):
+        return Quantity(1.0, (self.dimensionality))
+    @units.setter
+    def units(self, units):
         try:
             assert not isinstance(self.base, Quantity)
         except AssertionError:
@@ -167,15 +193,23 @@ class Quantity(np.ndarray):
         mag = self.magnitude
         mag *= cf
         self._dimensionality = to_u.dimensionality
-    units = property(_get_units, _set_units)
 
-    def rescale(self, units):
+    def rescale(self, units=None, dtype=None):
         """
-        Return a copy of the quantity converted to the specified units
+        Return a copy of the quantity converted to the specified units.
+        If `units` is `None`, an attempt will be made to rescale the quantity
+        to preferred units (see `rescale_preferred`).
         """
+        if units is None:
+            try:
+                return self.rescale_preferred()
+            except Exception as e:
+                raise Exception('No argument passed to `.rescale` and %s' % e)
         to_dims = validate_dimensionality(units)
+        if dtype is None:
+            dtype = self.dtype
         if self.dimensionality == to_dims:
-            return self.astype(self.dtype)
+            return self.astype(dtype)
         to_u = Quantity(1.0, to_dims)
         from_u = Quantity(1.0, self.dimensionality)
         try:
@@ -185,19 +219,39 @@ class Quantity(np.ndarray):
                 'Unable to convert between units of "%s" and "%s"'
                 %(from_u._dimensionality, to_u._dimensionality)
             )
-        return Quantity(cf*self.magnitude, to_u)
+        return Quantity(cf*self.magnitude, to_u, dtype=dtype)
+
+    def rescale_preferred(self):
+        """
+        Return a copy of the quantity converted to the preferred units and scale.
+        These will be identified from among the compatible units specified in the
+        list PREFERRED in this module. For example, a voltage quantity might be
+        converted to `mV`:
+        ```
+        import quantities as pq
+        pq.quantity.PREFERRED = [pq.mV, pq.pA]
+        old = 3.1415 * pq.V
+        new = old.rescale_preferred() # `new` will be 3141.5 mV.
+        ```
+        """
+        units_str = str(self.simplified.dimensionality)
+        for preferred in PREFERRED:
+            if units_str == str(preferred.simplified.dimensionality):
+                return self.rescale(preferred)
+        raise Exception("Preferred units for '%s' (or equivalent) not specified in "
+                        "quantites.quantity.PREFERRED." % self.dimensionality)
 
     @with_doc(np.ndarray.astype)
-    def astype(self, dtype=None):
+    def astype(self, dtype=None, **kwargs):
         '''Scalars are returned as scalar Quantity arrays.'''
-        ret = super(Quantity, self).astype(dtype)
+        ret = super(Quantity, self.view(Quantity)).astype(dtype, **kwargs)
         # scalar quantities get converted to plain numbers, so we fix it
         # might be related to numpy ticket # 826
         if not isinstance(ret, type(self)):
             if self.__array_priority__ >= Quantity.__array_priority__:
-                ret = type(self)(ret, self._dimensionality)
+                ret = type(self)(ret, self._dimensionality, dtype=self.dtype)
             else:
-                ret = Quantity(ret, self._dimensionality)
+                ret = Quantity(ret, self._dimensionality, dtype=self.dtype)
 
         return ret
 
@@ -206,24 +260,26 @@ class Quantity(np.ndarray):
 
     def __array_prepare__(self, obj, context=None):
         if self.__array_priority__ >= Quantity.__array_priority__:
-            result = obj.view(type(self))
+            res = obj if isinstance(obj, type(self)) else obj.view(type(self))
         else:
             # don't want a UnitQuantity
-            result = obj.view(Quantity)
+            res = obj.view(Quantity)
         if context is None:
-            return result
+            return res
 
         uf, objs, huh = context
         if uf.__name__.startswith('is'):
             return obj
-#        print self, obj, result, uf, objs
+        #print self, obj, res, uf, objs
         try:
-            result._dimensionality = p_dict[uf](*objs)
+            res._dimensionality = p_dict[uf](*objs)
         except KeyError:
-            print 'ufunc %r not supported by quantities' % uf
-            print 'please file a bug report by visiting'
-            print 'https://bugs.launchpad.net/python-quantities'
-        return result
+            raise ValueError(
+                """ufunc %r not supported by quantities
+                please file a bug report at https://github.com/python-quantities
+                """ % uf
+                )
+        return res
 
     def __array_wrap__(self, obj, context=None):
         if not isinstance(obj, Quantity):
@@ -234,58 +290,78 @@ class Quantity(np.ndarray):
     @with_doc(np.ndarray.__add__)
     @scale_other_units
     def __add__(self, other):
-        return super(Quantity, self).__add__(other)
+        return super().__add__(other)
 
     @with_doc(np.ndarray.__radd__)
     @scale_other_units
     def __radd__(self, other):
-        return super(Quantity, self).__radd__(other)
+        return np.add(other, self)
+        return super().__radd__(other)
 
     @with_doc(np.ndarray.__iadd__)
     @scale_other_units
     def __iadd__(self, other):
-        return super(Quantity, self).__iadd__(other)
+        return super().__iadd__(other)
 
     @with_doc(np.ndarray.__sub__)
     @scale_other_units
     def __sub__(self, other):
-        return super(Quantity, self).__sub__(other)
+        return super().__sub__(other)
 
     @with_doc(np.ndarray.__rsub__)
     @scale_other_units
     def __rsub__(self, other):
-        return super(Quantity, self).__rsub__(other)
+        return np.subtract(other, self)
+        return super().__rsub__(other)
 
     @with_doc(np.ndarray.__isub__)
     @scale_other_units
     def __isub__(self, other):
-        return super(Quantity, self).__isub__(other)
+        return super().__isub__(other)
+
+    @with_doc(np.ndarray.__mod__)
+    @scale_other_units
+    def __mod__(self, other):
+        return super().__mod__(other)
+
+    @with_doc(np.ndarray.__imod__)
+    @scale_other_units
+    def __imod__(self, other):
+        return super().__imod__(other)
 
     @with_doc(np.ndarray.__imul__)
     @protected_multiplication
     def __imul__(self, other):
-        return super(Quantity, self).__imul__(other)
+        return super().__imul__(other)
+
+    @with_doc(np.ndarray.__rmul__)
+    def __rmul__(self, other):
+        return np.multiply(other, self)
+        return super().__rmul__(other)
 
     @with_doc(np.ndarray.__itruediv__)
     @protected_multiplication
     def __itruediv__(self, other):
-        return super(Quantity, self).__itruediv__(other)
+        return super().__itruediv__(other)
 
-    @with_doc(np.ndarray.__idiv__)
-    @protected_multiplication
-    def __idiv__(self, other):
-        return super(Quantity, self).__itruediv__(other)
+    @with_doc(np.ndarray.__rtruediv__)
+    def __rtruediv__(self, other):
+        return np.true_divide(other, self)
+        return super().__rtruediv__(other)
 
     @with_doc(np.ndarray.__pow__)
     @check_uniform
     def __pow__(self, other):
-        return super(Quantity, self).__pow__(other)
+        return np.power(self, other)
 
     @with_doc(np.ndarray.__ipow__)
     @check_uniform
     @protected_power
     def __ipow__(self, other):
-        return super(Quantity, self).__ipow__(other)
+        return super().__ipow__(other)
+
+    def __round__(self, decimals=0):
+        return np.around(self, decimals)
 
     @with_doc(np.ndarray.__repr__)
     def __repr__(self):
@@ -301,21 +377,33 @@ class Quantity(np.ndarray):
             dims = self.dimensionality.string
         return '%s %s'%(str(self.magnitude), dims)
 
+    if tuple(map(int, np.__version__.split('.')[:2])) >= (1, 14):
+        # in numpy 1.14 the formatting of scalar values was changed
+        # see https://github.com/numpy/numpy/pull/9883
+
+        def __format__(self, format_spec):
+            ret = super().__format__(format_spec)
+            if self.ndim:
+                return ret
+            return ret + f' {self.dimensionality}'
+
     @with_doc(np.ndarray.__getitem__)
     def __getitem__(self, key):
-        if isinstance(key, int):
-            # This might be resolved by issue # 826
-            return Quantity(self.magnitude[key], self._dimensionality)
+        ret = super().__getitem__(key)
+        if isinstance(ret, Quantity):
+            return ret
         else:
-            return super(Quantity, self).__getitem__(key)
+            return Quantity(ret, self._dimensionality)
 
     @with_doc(np.ndarray.__setitem__)
     def __setitem__(self, key, value):
         if not isinstance(value, Quantity):
             value = Quantity(value)
-
-        # TODO: do we want this kind of magic?
-        self.magnitude[key] = value.rescale(self._dimensionality).magnitude
+        if self._dimensionality != value._dimensionality:
+            # Setting `dtype` to 'd' is done to ensure backwards
+            # compatibility, arguably it's questionable design.
+            value = value.rescale(self._dimensionality, dtype='d')
+        self.magnitude[key] = value
 
     @with_doc(np.ndarray.__lt__)
     @wrap_comparison
@@ -362,7 +450,10 @@ class Quantity(np.ndarray):
         #first get a dummy array from the ndarray method
         work_list = self.magnitude.tolist()
         #now go through and replace all numbers with the appropriate Quantity
-        self._tolist(work_list)
+        if isinstance(work_list, list):
+            self._tolist(work_list)
+        else:
+            work_list = Quantity(work_list, self.dimensionality)
         return work_list
 
     def _tolist(self, work_list):
@@ -380,35 +471,45 @@ class Quantity(np.ndarray):
 
     @with_doc(np.ndarray.sum)
     def sum(self, axis=None, dtype=None, out=None):
+        ret = self.magnitude.sum(axis, dtype, None if out is None else out.magnitude)
+        dim = self.dimensionality
+        if out is None:
+            return Quantity(ret, dim, copy=False)
+        if not isinstance(out, Quantity):
+            raise TypeError("out parameter must be a Quantity")
+        out._dimensionality = dim
+        return out
+
+    @with_doc(np.nansum)
+    def nansum(self, axis=None, dtype=None, out=None):
+        import numpy as np
         return Quantity(
-            self.magnitude.sum(axis, dtype, out),
+            np.nansum(self.magnitude, axis, dtype, out),
             self.dimensionality,
             copy=False
         )
 
     @with_doc(np.ndarray.fill)
-    def fill(self, scalar):
-        if not isinstance (scalar, Quantity):
-            scalar = Quantity(scalar, copy=False)
-
-        if scalar._dimensionality == self._dimensionality:
-            self.magnitude.fill(scalar.magnitude)
-        else:
-            raise ValueError("scalar must have the same units as self")
+    def fill(self, value):
+        self.magnitude.fill(value)
+        try:
+            self._dimensionality = value.dimensionality
+        except AttributeError:
+            pass
 
     @with_doc(np.ndarray.put)
-    def put(self, indicies, values, mode='raise'):
+    def put(self, indicies, values, mode='raise', dtype='d'):
         """
         performs the equivalent of ndarray.put() but enforces units
         values - must be an Quantity with the same units as self
         """
-        if isinstance(values, Quantity):
-            if values._dimensionality == self._dimensionality:
-                self.magnitude.put(indicies, values, mode)
-            else:
-                raise ValueError("values must have the same units as self")
-        else:
-            raise TypeError("values must be a Quantity")
+        # The default of `dtype` is set to 'd' to ensure backwards
+        # compatibility, arguably it's questionable design.
+        if not isinstance(values, Quantity):
+            values = Quantity(values)
+        if values._dimensionality != self._dimensionality:
+            values = values.rescale(self.units, dtype=dtype)
+        self.magnitude.put(indicies, values, mode)
 
     # choose does not function correctly, and it is not clear
     # how it would function, so for now it will not be implemented
@@ -433,31 +534,68 @@ class Quantity(np.ndarray):
 
     @with_doc(np.ndarray.max)
     def max(self, axis=None, out=None):
+        ret = self.magnitude.max(axis, None if out is None else out.magnitude)
+        dim = self.dimensionality
+        if out is None:
+            return Quantity(ret, dim, copy=False)
+        if not isinstance(out, Quantity):
+            raise TypeError("out parameter must be a Quantity")
+        out._dimensionality = dim
+        return out
+
+    @with_doc(np.ndarray.argmax)
+    def argmax(self, axis=None, out=None):
+        return self.magnitude.argmax(axis, out)
+
+    @with_doc(np.nanmax)
+    def nanmax(self, axis=None, out=None):
         return Quantity(
-            self.magnitude.max(),
+            np.nanmax(self.magnitude),
             self.dimensionality,
             copy=False
         )
 
     @with_doc(np.ndarray.min)
     def min(self, axis=None, out=None):
+        ret = self.magnitude.min(axis, None if out is None else out.magnitude)
+        dim = self.dimensionality
+        if out is None:
+            return Quantity(ret, dim, copy=False)
+        if not isinstance(out, Quantity):
+            raise TypeError("out parameter must be a Quantity")
+        out._dimensionality = dim
+        return out
+
+    @with_doc(np.nanmin)
+    def nanmin(self, axis=None, out=None):
         return Quantity(
-            self.magnitude.min(),
+            np.nanmin(self.magnitude),
             self.dimensionality,
             copy=False
         )
 
     @with_doc(np.ndarray.argmin)
-    def argmin(self,axis=None, out=None):
-        return self.magnitude.argmin()
+    def argmin(self, axis=None, out=None):
+        return self.magnitude.argmin(axis, out)
+
+    @with_doc(np.nanargmin)
+    def nanargmin(self,axis=None, out=None):
+        return np.nanargmin(self.magnitude)
+
+    @with_doc(np.nanargmax)
+    def nanargmax(self,axis=None, out=None):
+        return np.nanargmax(self.magnitude)
 
     @with_doc(np.ndarray.ptp)
     def ptp(self, axis=None, out=None):
-        return Quantity(
-            self.magnitude.ptp(),
-            self.dimensionality,
-            copy=False
-        )
+        ret = self.magnitude.ptp(axis, None if out is None else out.magnitude)
+        dim = self.dimensionality
+        if out is None:
+            return Quantity(ret, dim, copy=False)
+        if not isinstance(out, Quantity):
+            raise TypeError("out parameter must be a Quantity")
+        out._dimensionality = dim
+        return out
 
     @with_doc(np.ndarray.clip)
     def clip(self, min=None, max=None, out=None):
@@ -478,43 +616,89 @@ class Quantity(np.ndarray):
             max.rescale(self._dimensionality).magnitude,
             out
         )
-        return Quantity(clipped, self.dimensionality, copy=False)
+        dim = self.dimensionality
+        if out is None:
+            return Quantity(clipped, dim, copy=False)
+        if not isinstance(out, Quantity):
+            raise TypeError("out parameter must be a Quantity")
+        out._dimensionality = dim
+        return out
 
     @with_doc(np.ndarray.round)
     def round(self, decimals=0, out=None):
-        return Quantity(
-            self.magnitude.round(decimals, out),
-            self.dimensionality,
-            copy=False
-        )
+        ret = self.magnitude.round(decimals, None if out is None else out.magnitude)
+        dim = self.dimensionality
+        if out is None:
+            return Quantity(ret, dim, copy=False)
+        if not isinstance(out, Quantity):
+            raise TypeError("out parameter must be a Quantity")
+        out._dimensionality = dim
+        return out
 
     @with_doc(np.ndarray.trace)
     def trace(self, offset=0, axis1=0, axis2=1, dtype=None, out=None):
+        ret = self.magnitude.trace(offset, axis1, axis2, dtype, None if out is None else out.magnitude)
+        dim = self.dimensionality
+        if out is None:
+            return Quantity(ret, dim, copy=False)
+        if not isinstance(out, Quantity):
+            raise TypeError("out parameter must be a Quantity")
+        out._dimensionality = dim
+        return out
+
+    @with_doc(np.ndarray.squeeze)
+    def squeeze(self, axis=None):
         return Quantity(
-            self.magnitude.trace(offset, axis1, axis2, dtype, out),
+            self.magnitude.squeeze(axis),
             self.dimensionality,
             copy=False
         )
 
     @with_doc(np.ndarray.mean)
     def mean(self, axis=None, dtype=None, out=None):
+        ret = self.magnitude.mean(axis, dtype, None if out is None else out.magnitude)
+        dim = self.dimensionality
+        if out is None:
+            return Quantity(ret, dim, copy=False)
+        if not isinstance(out, Quantity):
+            raise TypeError("out parameter must be a Quantity")
+        out._dimensionality = dim
+        return out
+
+    @with_doc(np.nanmean)
+    def nanmean(self, axis=None, dtype=None, out=None):
+        import numpy as np
         return Quantity(
-            self.magnitude.mean(axis, dtype, out),
+            np.nanmean(self.magnitude, axis, dtype, out),
             self.dimensionality,
             copy=False)
 
     @with_doc(np.ndarray.var)
-    def var(self, axis=None, dtype=None, out=None):
-        return Quantity(
-            self.magnitude.var(axis, dtype, out),
-            self._dimensionality**2,
-            copy=False
-        )
+    def var(self, axis=None, dtype=None, out=None, ddof=0):
+        ret = self.magnitude.var(axis, dtype, out, ddof)
+        dim = self._dimensionality**2
+        if out is None:
+            return Quantity(ret, dim, copy=False)
+        if not isinstance(out, Quantity):
+            raise TypeError("out parameter must be a Quantity")
+        out._dimensionality = dim
+        return out
 
     @with_doc(np.ndarray.std)
-    def std(self, axis=None, dtype=None, out=None):
+    def std(self, axis=None, dtype=None, out=None, ddof=0):
+        ret = self.magnitude.std(axis, dtype, out, ddof)
+        dim = self.dimensionality
+        if out is None:
+            return Quantity(ret, dim, copy=False)
+        if not isinstance(out, Quantity):
+            raise TypeError("out parameter must be a Quantity")
+        out._dimensionality = dim
+        return out
+
+    @with_doc(np.nanstd)
+    def nanstd(self, axis=None, dtype=None, out=None, ddof=0):
         return Quantity(
-            self.magnitude.std(axis, dtype, out),
+            np.nanstd(self.magnitude, axis, dtype, out, ddof),
             self._dimensionality,
             copy=False
         )
@@ -526,11 +710,25 @@ class Quantity(np.ndarray):
         else:
             power = self.shape[axis]
 
-        return Quantity(
-            self.magnitude.prod(axis, dtype, out),
-            self._dimensionality**power,
-            copy=False
-        )
+        ret = self.magnitude.prod(axis, dtype, None if out is None else out.magnitude)
+        dim = self._dimensionality**power
+        if out is None:
+            return Quantity(ret, dim, copy=False)
+        if not isinstance(out, Quantity):
+            raise TypeError("out parameter must be a Quantity")
+        out._dimensionality = dim
+        return out
+
+    @with_doc(np.ndarray.cumsum)
+    def cumsum(self, axis=None, dtype=None, out=None):
+        ret = self.magnitude.cumsum(axis, dtype, None if out is None else out.magnitude)
+        dim = self.dimensionality
+        if out is None:
+            return Quantity(ret, dim, copy=False)
+        if not isinstance(out, Quantity):
+            raise TypeError("out parameter must be a Quantity")
+        out._dimensionality = dim
+        return out
 
     @with_doc(np.ndarray.cumprod)
     def cumprod(self, axis=None, dtype=None, out=None):
@@ -539,39 +737,36 @@ class Quantity(np.ndarray):
             raise ValueError(
                 "Quantity must be dimensionless, try using simplified"
             )
-        else:
-            return super(Quantity, self).cumprod(axis, dtype, out)
+
+        ret = self.magnitude.cumprod(axis, dtype, out)
+        dim = self.dimensionality
+        if out is None:
+            return Quantity(ret, dim, copy=False)
+        if isinstance(out, Quantity):
+            out._dimensionality = dim
+        return out
 
     # list of unsupported functions: [choose]
 
-    def __getstate__(self):
-        """
-        Return the internal state of the quantity, for pickling
-        purposes.
-
-        """
-        cf = 'CF'[self.flags.fnc]
-        state = (1,
-                 self.shape,
-                 self.dtype,
-                 self.flags.fnc,
-                 self.tostring(cf),
-                 self._dimensionality,
-                 )
-        return state
-
     def __setstate__(self, state):
-        (ver, shp, typ, isf, raw, units) = state
-        np.ndarray.__setstate__(self, (shp, typ, isf, raw))
+        ndarray_state = state[:-1]
+        units = state[-1]
+        np.ndarray.__setstate__(self, ndarray_state)
         self._dimensionality = units
 
     def __reduce__(self):
         """
         Return a tuple for pickling a Quantity.
         """
+        reconstruct,reconstruct_args,state = super().__reduce__()
+        state = state + (self._dimensionality,)
         return (_reconstruct_quantity,
                 (self.__class__, np.ndarray, (0, ), 'b', ),
-                self.__getstate__())
+                state)
+
+    def __deepcopy__(self, memo_dict):
+        # constructor copies by default
+        return Quantity(self.magnitude, self.dimensionality)
 
 
 def _reconstruct_quantity(subtype, baseclass, baseshape, basetype,):
